/*
 * Find and report all code similarities as exist in the ctfnode.db database.
 * Copyright (c) Warren Toomey, under the GPL license.
 * 
 * $Revision: 1.27 $
 */
#undef DEBUG
#undef STATS

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include "dbfiles.h"
#include "tokens.h"

char **namearray;		/* Array of CTF files in the database */
FILE **ctf_handle;		/* Array of open filehandles, one per CTF */
int numfileids;			/* Size of the above two arrays */

int threshold = 16;		/* Threshold above which we report a run */
int do_isomorph_comparison = 0;	/* Do an isomorphic comparison on the code */
int isomorph_count_threshold=6;	/* Only allow 3 isomorphic relations */
int showtokens = 0;		/* If 1, print out tokens in a run */
int crosscompare = 0;		/* If 1, print out actual code in a run */
int side_by_side = 0;		/* Do we display the output side by side? */
#ifdef STATS
int coll_count=0;		/* Count the # of collions found in the db */
#endif

/* Isomorph tables */
/* We need two tables: one to record the match from a dst value */
/* to a src value, and another to record the src to dst value match */
uint16_t isodtos[65536];		/* Destination to source isomorphism */
uint16_t isostod[65536];		/* Source to destination isomorphism */
uint16_t isoseen[65536];		/* =1 if we have seen this id value */
int max_isoseen=0;			/* Number of relationships seen */


/*
 * We keep a linked list of confirmed token runs found so far,
 * to ensure that we don't try to re-evaluate them.
 */
struct run {
  Ctfhashnode src;		/* Details of the src run start, line start */
  Ctfhashnode dst;		/* Details of the dst run start, line start */
  uint32_t src_end_offset;	/* End offset of the run in the src tree */
  uint32_t dst_end_offset;	/* End offset of the run in the dst tree */
  uint32_t runlength;		/* Length of the run in tokens */
  struct run *next;		/* Next node in the list */
};

struct run *Runhead = NULL;	/* Head of the run tree */

/*
 * Before we can insert a node, we have to determine if it is a subset of
 * a run already in the list. If so, exit. Alternatively, the one in the list
 * is the smaller one, so replace it. Alternatively, no subset, so insert it.
 */
struct run * add_run(Ctfhashnode * src, Ctfhashnode * dst, uint32_t length,
	uint32_t srcend, uint32_t dstend)
{
  struct run *this;

  /* Ensure that the src information is for the smallest fileid.
   * This means that we don't have to compare in 2 directions.
   * This also makes our printout of the runs nicer later.
   */
  if (dst->fileid < src->fileid) {
    Ctfhashnode *tmpnode = src; src = dst; dst = tmpnode;
    uint32_t tempend = srcend; srcend = dstend; dstend = tempend;
  }

  /* Walk the existing list, looking for possible overlaps */
  for (this = Runhead; this != NULL; this = this->next) {

    /* Only compare the same fileids */
    if ((this->src.fileid == src->fileid) && (this->dst.fileid == dst->fileid))
    {

      /* If our params are a subset of this run, return */
      if ((src->offset >= this->src.offset) &&
	  (srcend <= this->src_end_offset) &&
	  (dst->offset >= this->dst.offset) &&
	  (dstend <= this->dst_end_offset)) {
	return(this);
      }

      /* If this run is a subset of the param run, replace this */
      if ((this->src.offset >= src->offset) &&
	  (this->src_end_offset <= srcend) &&
	  (this->dst.offset >= dst->offset) &&
	  (this->dst_end_offset <= dstend)) {
        memcpy(&(this->src), src, sizeof(Ctfhashnode));
        memcpy(&(this->dst), dst, sizeof(Ctfhashnode));
	this->src_end_offset = srcend;
	this->dst_end_offset = dstend;
	this->runlength = length;
	return(this);
      }
    }
  }

  /* No subset, so we can add a new node to the list */
  struct run *runnode = malloc(sizeof(struct run));
  if (runnode == NULL) return(NULL);

  /* Copy the details of the run from the inputs */
  memcpy(&(runnode->src), src, sizeof(Ctfhashnode));
  memcpy(&(runnode->dst), dst, sizeof(Ctfhashnode));
  runnode->src_end_offset = srcend;
  runnode->dst_end_offset = dstend;
  runnode->runlength = length;

  /* Insert the new node at the head of the list. */
  runnode->next = Runhead;
  Runhead = runnode;
  return(runnode);
}

/* Given a run, print out the tokens in the run
 * in much the same way as we do in detok.c
 */
void print_tokens(struct run *node)
{
  uint32_t val;
  unsigned int ch;
  int length= node->runlength;
  int fid= node->src.fileid;
  uint32_t offset= node->src.offset;
  uint32_t line= node->src.line_number;

  printf("%5d:   ", line);
  while ((length >= 0) &&
	 ((ch = get_token(ctf_handle[fid], &offset, &val, NULL)) != 0)) {
    switch (ch) {
    case FILENAME:
      line = 1;
      break;
    case LINE:
      line++;
      break;
    default:
      length--;
    }
    print_token(ch, line, val, "");
  }
  printf("\n\n");
}

/*
 * Print out lines of two files side-by-side. This is best viewed with a
 * terminal window 160 characters wide! If neither file exists, the function
 * simply returns with no error message.
 */
void paste_files(char *file1, char *file2, struct run *node)
{
  FILE *f1in, *f2in;
  char buf[1024];
  int count_to_eighty = 0;
  int i, bptr, maxlines;
  int numlines1, numlines2;
  int tab_upto = 0;
  int side_side= side_by_side;		/* This may change */
  int start1= node->src.line_number;
  int end1= node->src.last_line_number;
  int start2= node->dst.line_number;
  int end2= node->dst.last_line_number;

  f1in = fopen(file1, "r");
  if (f1in == NULL)
    side_side=0;

  f2in = fopen(file2, "r");
  if (f2in == NULL)
    side_side=0;

  /* We can't display either file, simply show the tokens */
  if ((f1in==NULL) && (f2in==NULL)) {
    print_tokens(node);
    return;
  }

  /* Get maximum number of lines */
  numlines1= end1 - start1 + 1;
  numlines2= end2 - start2 + 1;
  maxlines = (numlines1 > numlines2) ? numlines1 : numlines2;
  start1--;
  start2--;

  /* Read up to line start1 in file1 */
  if (f1in!=NULL)
    while (start1--)
      fgets(buf, 1020, f1in);

  /* Read up to line start2 in file2 */
  if (f2in!=NULL)
    while (start2--)
      fgets(buf, 1020, f2in);

  if (side_side) {
    /* Now print out the lines */
    for (i = 0; i < maxlines; i++, numlines1--, numlines2--) {

      /* Print the left-hand line */
      if (numlines1 > 0) {
	fgets(buf, 1020, f1in);
	buf[strlen(buf) - 1] = '\0';	/* Remove newline */

	/*
	 * I used to printf("%-80s",buf); here, but it never did what I
	 * wanted, so I now print out exactly 80 chars by hand. This is
	 * because we have to deal with tabs, dammit!
	 */
	tab_upto = 0;
	bptr = 0;
	for (count_to_eighty = 0; count_to_eighty < 80; count_to_eighty++) {
	  if (tab_upto > count_to_eighty) {
	    fputc(' ', stdout);
	    continue;
	  }
	  if (buf[bptr] == '\0')
	    break;
	  if (buf[bptr] == '\t') {

	    /* Calculate the next highest multiple of 8 */
	    tab_upto = (count_to_eighty + 8) & ~7;
	    if (tab_upto >= 80)
	      break;
	    fputc(' ', stdout);
	    bptr++;
	    continue;
	  }
	  fputc(buf[bptr++], stdout);
	}
	for (; count_to_eighty < 80; count_to_eighty++) {
	  fputc(' ', stdout);
	}
	printf(" ");
      } else
	printf("%81s", " ");

      /* Print the right-hand line */
      if (numlines2 > 0) {
	fgets(buf, 1020, f2in);
	fputs(buf, stdout);
      } else
	printf("\n");
    }
  } else {			/* Not side-by-side */
    if (f1in!=NULL)
      for (i = 0; i < numlines1; i++) {
        fgets(buf, 1020, f1in);
        fputs(buf, stdout);
      }
    else
      print_tokens(node);
    printf("=====================================\n");
    if (f2in!=NULL)
      for (i = 0; i < numlines2; i++) {
        fgets(buf, 1020, f2in);
        fputs(buf, stdout);
      }
    else
      print_tokens(node);
  }
  printf("\n");
  if (numlines1 > numlines2)
    printf("\n");
  if (f1in!=NULL) fclose(f1in);
  if (f2in!=NULL) fclose(f2in);
}

/* Given a run, print out the 3-column header.
 * Then determine if and how to print out
 * the details of the run.
 */
void print_listrun(struct run *node)
{
  char **srcname, **dstname;
  uint32_t timestamp, off;
  srcname = (char **)malloc(sizeof(char *));
  dstname = (char **)malloc(sizeof(char *));
  off= node->src.name_offset;
  get_token(ctf_handle[node->src.fileid], &off, &timestamp, srcname);
  off= node->dst.name_offset;
  get_token(ctf_handle[node->dst.fileid], &off, &timestamp, dstname);

  printf("%d  %s:%d-%d  %s:%d-%d\n", node->runlength, *srcname,
	   node->src.line_number, node->src.last_line_number, *dstname,
	   node->dst.line_number, node->dst.last_line_number);

  if (showtokens)
    print_tokens(node);
  if (crosscompare)
    paste_files(*srcname, *dstname, node);
  free(*srcname);
  free(srcname);
  free(*dstname);
  free(dstname);
}

/* Print out all of the runs. */
void print_listruns(void)
{
  struct run *this;

  for (this = Runhead; this != NULL; this = this->next)
    print_listrun(this);
}

/*
 * Given two Ctfhashnodes, do a full comparison
 * and return the run length found.
 */
int compare_runs(Ctfhashnode * src, Ctfhashnode * dst,
	     				uint32_t * srcend, uint32_t * dstend)
{
  int runlength = 0;
  uint32_t srcfid = src->fileid;
  uint32_t dstfid = dst->fileid;
  uint32_t srcoffset = src->offset;
  uint32_t dstoffset = dst->offset;
  uint32_t srcendline = src->line_number;
  uint32_t dstendline = dst->line_number;
  uint32_t nextsrcoff, nextsrcline;
  uint32_t nextdstoff, nextdstline;
  uint8_t srctoken, dsttoken;
  uint32_t srcid, dstid;

#define MYASSERT
#ifdef MYASSERT
  if ((srcfid > numfileids) || (dstfid > numfileids)) {
    printf("One of the fids is out of bounds!\n");
    exit(1);
  }
#endif

/*
 * Here are some comments on the isomorph[] arrays. We would like to identify
 * code fragments that are similar, even if variables have been renamed or
 * even declared in a different order. For example, the following two
 * functions are essentially identical.
 * 
 * int maxofthree(int x, int y, int z)    int bigtriple(int a, int b, int c)
 * {                                      {
 *   if ((x>y) && (x>z)) return(x);          if ((a>b) && (a>c)) return(a);
 *   if (y>z) return(y);                     if (b>c) return(b);
 *   return(z);                              return(c);
 * }                                      }
 * 
 * The tokenised forms of these functions are obviously different.
 * However, it is possible to compare these and show that they are identical.
 * If we start with the first "if" statement, a straight lexical comparison
 * shows that they are different. But at the first difference
 * (id1123 cf. id456), we put an entry into the isodtos[] array so that
 * isodtos[456]= 123, i.e. the identifier #456 on the RHS is isomorphic to
 * identifier #123 on the LHS, and vice versa. Thus when we get to the first
 * "return" statement, we can see that they are also isomorphic. We can thus
 * conclude that everything from the "if" down to the "}" is isomorphic and
 * thus identical.
 */

 /* Clear the identifer isomorph table for this run.
  * I used to simply memset() both isodtos[] and
  * isostod[], but it is faster to track and delete
  * only those entries that we used.
  */
  if (do_isomorph_comparison) {
    while (--max_isoseen>=0) {
      isodtos[isoseen[max_isoseen]]= 0;
      isostod[isoseen[max_isoseen]]= 0;
    }
    max_isoseen=0;
  }

  /* Set up the next variables, because we have to fall back to
   * the previous values when there is a match.
   */
  nextsrcoff= srcoffset;   nextdstoff= dstoffset;
  nextsrcline= srcendline; nextdstline= dstendline;

  /*
   * We know that the first TUPLE_SIZE tokens match and the XOR. Now do a
   * full match from the start offsets in both until something fails.
   */
  while (1) {

    /* Get the next token from each file */
    srctoken = get_nonltoken(ctf_handle[srcfid], &nextsrcoff,
			     &srcid, &nextsrcline);
    dsttoken = get_nonltoken(ctf_handle[dstfid], &nextdstoff,
			     &dstid, &nextdstline);

    /* Stop if no match, or if we hit a new file */
    if ((srctoken == FILENAME) || (dsttoken == FILENAME) ||
        (srctoken != dsttoken))
      break;

    /* Short-circuit the isomorph test if both ids the same */
    if (srcid==dstid)
      goto loopback2;		/* Yuk, a goto! */

    /* If we hit an IDENTIFIER, do an isomorphic comparison */
    if (do_isomorph_comparison && (srctoken==IDENTIFIER)) {

      /* First thing, record a mapping each way if there is none */
      if (isodtos[dstid] == 0) {
        isodtos[dstid] = srcid;
        isoseen[max_isoseen++]= dstid;
      }
      if (isostod[srcid] == 0) {
        isostod[srcid] = dstid;
        isoseen[max_isoseen++]= srcid;
      }

      /* Now reject if the mappings fail in either direction */
      if (isodtos[dstid] != srcid)
        break;
      if (isostod[srcid] != dstid)
        break;

      /* Stop now if we have reach the threshold on the number
       * of isomorphic relations that we can have.
       * isomorph_count_threshold is always doubled
       * because we always have a 2-way relation.
       */
      if (max_isoseen >= isomorph_count_threshold)
	break;
    } else {
      /* Not isomorphic and/or not IDENTIFIER, any id mismatch ends the run */
      if (srcid != dstid)
        break;
    }
    
    /* If we got here, there was one more token+value match.
     * Update the offsets, linenumbers and the run length.
     */
loopback2:
    runlength++;
    srcoffset= nextsrcoff;
    dstoffset= nextdstoff;
    srcendline= nextsrcline;
    dstendline= nextdstline;
  }

  /* Update the end offsets and end line numbers */
  *srcend = srcoffset;
  *dstend = dstoffset;
  src->last_line_number = srcendline;
  dst->last_line_number = dstendline;
  return (runlength);
}

void usage(void)
{
  fprintf(stderr,
  "Usage: ctcompare [-t] [-x] [-s] [-i] [-I nnn] [-n nnn] [file1.ctf] [file2.ctf]\n");
  fprintf(stderr, "\t-t: show matching tokens when match is found\n");
  fprintf(stderr, "\t-x: show matching source lines when match is found\n");
  fprintf(stderr, "\t-s: print results side by side on same line, implies -x\n");
  fprintf(stderr, "\t-i: enable isomorphic code comparison\n");
  fprintf(stderr, "\t-I nnn: limit the # isomorphic relations to nnn\n");
  fprintf(stderr, "\t-n nnn: set the minimum matching run length to nnn\n");
  fprintf(stderr, "    If file1.ctf or both file1.ctf & file2.ctf specified,\n");
  fprintf(stderr, "    only this/these files will be compared against.\n");
  exit(1);
}

int main(int argc, char *argv[])
{
  uint8_t *key, *oldkey;	/* Key retrieved */
  int numvals;
  int ch, i, j, firstfile;
  int file1_id = 0, file2_id = 0;
  int has_many_files;
  uint32_t runlength;
  Ctfhashnode *node;
  struct run *runnode;
  uint32_t srcendoffset, dstendoffset;

  /* Process options */
  while ((ch = getopt(argc, argv, "iI:tsxn:")) != -1) {
    switch (ch) {
    case 'I':
      /* isomorph_count_threshold is always doubled
       * because we always have a 2-way relation.
       */
      i = 2 * atoi(optarg);
      if (i < 2) {
	fprintf(stderr, "Bad value for -I, must be 1 or greater\n");
      } else
	isomorph_count_threshold=i;
    case 'i':
      do_isomorph_comparison = 1;
      break;
    case 't':
      if (crosscompare) {
	fprintf(stderr, "-t and -x are mutually exclusive\n");
	crosscompare = 0;
      }
      showtokens = 1;
      break;
    case 's':
      side_by_side = 1;
    case 'x':
      if (showtokens) {
	fprintf(stderr, "-t and -x are mutually exclusive\n");
	showtokens = 0;
      }
      crosscompare = 1;
      break;
    case 'n':
      i = atoi(optarg);
      if (i < 16) {
	fprintf(stderr, "Bad value for -n, must be 16 or greater\n");
      } else
	threshold = i;
      break;
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (argc > 2)
    usage();

  /*
   * Get the names and fileids of the CTF files in the database.
   */
  numfileids = get_fileidnames(&namearray);
  ctf_handle = (FILE **) malloc(numfileids * sizeof(FILE *));
  if (ctf_handle == NULL) {
    fprintf(stderr, "Cannot allocate %d file handles\n", numfileids);
    exit(1);
  }

  /*
   * Open up all of the CTF files.
   */
  for (i = 1; i < numfileids; i++) {
    ctf_handle[i] = fopen(namearray[i], "r");
    if (ctf_handle[i] == NULL) {
      fprintf(stderr, "Cannot open %s for reading\n", namearray[i]);
      exit(1);
    }
  }

  /* Determine if any ctf filter files exists in the database. */
  if (argc > 0) {
    for (i = 1; i < numfileids; i++) {
      if (!strcmp(argv[0], namearray[i])) {
	file1_id = i;
	break;
      }
    }
    if (file1_id == 0) {
      fprintf(stderr, "%s not represented in the database\n", argv[0]);
      exit(1);
    }
  }

  if (argc > 1) {
    for (i = 1; i < numfileids; i++) {
      if (!strcmp(argv[1], namearray[i])) {
	file2_id = i;
	break;
      }
    }
    if (file2_id == file1_id) {
      fprintf(stderr, "%s not represented in the database\n", argv[1]);
      exit(1);
    }
  }

  /* Get first key */
  if (get_firstkey(&key, &node, &numvals)==0) {
    printf("Cannot get first key\n"); exit(1);
  }

  /* Loop while we have a key */
  while (1) {

    /* Skip values which are size 1 */
    if (numvals < 2)
      goto loopback;		/* Yuk, a goto! */

    /* Skip values which are all for the same CTF file */
    has_many_files = 0;
    firstfile = node[0].fileid;
    for (i = 0; i < numvals; i++) {
      if (firstfile != node[i].fileid) {
	has_many_files = 1;
	break;
      }
    }

    if (has_many_files == 0)
      goto loopback;		/* Yuk, a goto! */

    /* Time to do a full run length analysis of the combinations */
    for (i = 0; i < numvals; i++) {

      /* Skip if both filter files on and no match against either */
      if (file1_id && file2_id
	  && (node[i].fileid != file1_id) && (node[i].fileid != file2_id))
	continue;

      for (j = i; j < numvals; j++) {

	/* Don't compare within the same CTF file */
	if (node[i].fileid == node[j].fileid)
	  continue;

	/* If no file1_id match, and none this time, skip */
	if (file1_id && (node[i].fileid != file1_id) &&
	   		(node[j].fileid != file1_id))
	  continue;

	/* Check against file2_id if it exists */
	if (file2_id) {
	  /*
	   * If there was a prev match against file1_id, skip if no
	   * file2_match
	   */
	  if ((node[i].fileid == file1_id) && (node[j].fileid != file2_id))
	    continue;
	  /* And vice versa */
	  if ((node[i].fileid == file2_id) && (node[j].fileid != file1_id))
	    continue;
	}
#ifdef STATS
	coll_count++;
#endif
	runlength = compare_runs(&node[i], &node[j],
						&srcendoffset, &dstendoffset);
	if (runlength >= threshold) {
	  /* Save this run into the linked list of known runs */
	  runnode=
	    add_run(&node[i], &node[j], runlength, srcendoffset, dstendoffset);
#ifdef DEBUG
	  if (runnode) {
	    printf("Debug "); print_listrun(runnode);
	  }
#endif
	}
      }
    }

loopback:
#ifdef GDBM
    /* Free the key/value retrieved & get the next one */
    free(node);
#endif
    oldkey= key;
    if (get_nextkey(oldkey, &key, &node, &numvals)==0)
      break;
#ifdef GDBM
    free(oldkey);
#endif
  }

  /* All comparions done, print out the final list of runs */
#ifdef STATS
  printf("Number of collisions tested: %d\n", coll_count);
#endif
  print_listruns();
  return (0);
}
